#pragma once

#include <cstdint>
#include <map>
#include <memory>

#include <amdis/AdaptionInterface.hpp>
#include <amdis/GridTransfer.hpp>
#include <amdis/common/ConcurrentCache.hpp>

namespace AMDiS
{
  /**
   * \addtogroup Adaption
   * @{
   **/

  /// Static administration class for automatic handling of DOFVectors during grid adaption
  class GridTransferManager
  {
    using Self = GridTransferManager;
    using Key = std::uintptr_t;
    using Data = std::shared_ptr<AdaptionInterface>;

  private:
    // Constructors. Since this is a static class you cannot call any constructor.
    GridTransferManager() = delete;

  public:
    /**
     * \brief Adapts the grid according to previously set marks and performs data transfer on
     *        all DOFVectors associated to the given grid.
     *
     * This function retrieves the grid's stored GridTransfer and calls its adaptation methods.
     * During this process the grid will change according to previously set grid element markings.
     * All DOFVectors associated to the given grid will be interpolated to the new grid according to
     * their DataTransfer member object.
     *
     * Takes a grid as argument. Marking of the grid needs to be performed prior to calling this.
     * Returns true if the grid changed during adaptation.
     **/
    template <class Grid>
    static bool adapt(Grid& grid)
    {
      auto transfer = gridTransfer(grid);
      bool adapted = transfer->preAdapt();
      adapted |= transfer->adapt();
      transfer->postAdapt();
      return adapted;
    }

    /// Returns the GridTransfer bound to a Grid, to be used during the adaptation cycle
    template <class Grid>
    static std::shared_ptr<GridTransfer<Grid>> gridTransfer(Grid& grid)
    {
      auto transfer = get(grid);
      transfer->bind(grid);
      return transfer;
    }

  public:
    template <class Grid, class Data>
    static void attach(Grid const& grid, Data& data)
    {
      get(grid)->attach(data);
    }

    template <class Grid, class Data>
    static void detach(Grid const& grid, Data& data)
    {
      get(grid)->detach(data);
    }


  private:
    /// \brief Returns the GridTransfer class, used for attaching and detaching Data
    /**
     * NOTE: The returned GridTransfer is not yet bound to a grid, since a
     * mutable reference `Grid&` is needed for the binding.
     **/
    template <class Grid>
    static std::shared_ptr<GridTransfer<Grid>> get(Grid const& grid)
    {
      Key key = Key(&grid);
      GridTransferCache cache;
      auto& gridTransferInterface = cache.get(key, [&](Key const&)
      {
        return std::make_unique<GridTransfer<Grid>>();
      });

      return std::dynamic_pointer_cast<GridTransfer<Grid>>(gridTransferInterface);
    }

  private:
    // Associate to each Grid (represented as address) a GridTransfer
    using GridTransferCache = ConcurrentCache< Key, Data, StaticLockedPolicy, std::map<Key, Data> >;
  };

  /// @}

} // end namespace AMDiS
