#pragma once

#include <cmath>

#include <dune/common/timer.hh>

#include <amdis/OperatorList.hpp>
#include <amdis/common/Math.hpp>
#include <amdis/typetree/MultiIndex.hpp>
#include <amdis/typetree/TreePath.hpp>

namespace AMDiS
{
  /**
   * Basis implementation of DOFMatrix, i.e. a sparse matrix storing all the
   * assembled Operators indexed with DOF indices. The matrix data is associated
   * to a row and column global basis.
   *
   * \tparam RB  Basis of the matrix rows
   * \tparam CB  Basis of matrix columns
   * \tparam Backend  A linear-algebra backend for the matrix storage
   **/
  template <class RB, class CB, class Backend>
  class DOFMatrixBase
  {
  public:
    /// The type of the finite element space / basis of the row
    using RowBasis = RB;
    using RowLocalView = typename RowBasis::LocalView;

    /// The type of the finite element space / basis of the column
    using ColBasis = CB;
    using ColLocalView = typename ColBasis::LocalView;

    using Element = typename RowLocalView::Element;
    using Geometry = typename Element::Geometry;

    /// The index/size - type
    using size_type = typename RowBasis::size_type;
    using value_type = typename Backend::value_type;

    /// The type of the data matrix used in the backend
    using BaseMatrix = typename Backend::BaseMatrix;

    /// The type of the matrix filled on an element with local contributions
    using ElementMatrix = Dune::DynamicMatrix<value_type>;

  public:
    /// Constructor. Stores the shared_ptr to the bases.
    DOFMatrixBase(std::shared_ptr<RowBasis> rowBasis, std::shared_ptr<ColBasis> colBasis)
      : rowBasis_(rowBasis)
      , colBasis_(colBasis)
    {
      operators_.init(*rowBasis_, *colBasis_);
    }

    /// Constructor. Wraps the reference into a non-destroying shared_ptr or moves the basis into a new shared_ptr.
    template <class RB_, class CB_>
    DOFMatrixBase(RB_&& rowBasis, CB_&& colBasis)
      : DOFMatrixBase(Dune::wrap_or_move(FWD(rowBasis)), Dune::wrap_or_move(FWD(colBasis)))
    {}

    /// Return the row-basis \ref rowBasis of the matrix
    std::shared_ptr<RowBasis const> rowBasis() const
    {
      return rowBasis_;
    }

    /// Return the col-basis \ref colBasis of the matrix
    std::shared_ptr<ColBasis const> colBasis() const
    {
      return colBasis_;
    }

    /// Return the data-matrix
    BaseMatrix const& matrix() const
    {
      return backend_.matrix();
    }

    /// Return the data-matrix
    BaseMatrix& matrix()
    {
      return backend_.matrix();
    }

    /// Return the size of the \ref rowBasis_
    size_type rows() const
    {
      return rowBasis_->dimension();
    }

    /// Return the size of the \ref colBasis_
    size_type cols() const
    {
      return colBasis_->dimension();
    }

    /// Initialize the matrix for insertion, e.g. allocate the non-zero pattern
    /// If \p setToZero is true, the matrix is set to 0
    void init(bool asmMatrix);

    /// Finish the matrix insertion, e.g. cleanup or final insertion
    void finish(bool asmMatrix);

    /// Insert a block of values into the matrix (add to existing values)
    /// The global matrix indices are determined by the corresponding localviews.
    void insert(RowLocalView const& rowLocalView,
                ColLocalView const& colLocalView,
                ElementMatrix const& elementMatrix);

    /// Insert a single value into the matrix (add to existing value)
    template <class RowIndex, class ColIndex>
    void insert(RowIndex row, ColIndex col, typename Backend::value_type const& value)
    {
      backend_.insert(flatMultiIndex(row), flatMultiIndex(col), value);
    }

    /// \brief Associate a local operator with this DOFMatrix
    /**
     * Stores an operator in a list that gets assembled during a call to \ref assemble().
     * The operator may be assigned to a specific context, i.e. either an element
     * operator, an intersection operator, or a boundary operator.
     * The \p row and \p col tree paths specify the sub-basis for test and trial
     * functions the operator is applied to.
     *
     * \tparam ContextTag  One of \ref tag::element_operator, \ref tag::intersection_operator
     *                     or \ref tag::boundary_operator indicating where to assemble this operator.
     * \tparam Expr        An pre-operator that can be bound to a gridView, or a valid
     *                      GridOperator.
     * \tparam row  A tree-path for the RowBasis
     * \tparam col  A tree-path for the ColBasis
     *
     * [[expects: row is valid tree-path in RowBasis]]
     * [[expects: col is valid tree-path in ColBasis]]
     **/
    // TODO: add method without contextTag.
    template <class ContextTag, class Expr,
              class RowTreePath = RootTreePath, class ColTreePath = RootTreePath>
    void addOperator(ContextTag contextTag, Expr const& expr,
                     RowTreePath row = {}, ColTreePath col = {});

    /// Assemble the matrix operators on the bound element.
    void assemble(RowLocalView const& rowLocalView,
                  ColLocalView const& colLocalView);

    /// Assemble all matrix operators, TODO: incooperate boundary conditions
    void assemble();

    /// Number of nonzeros in the matrix
    size_type nnz() const
    {
      return backend_.nnz();
    }

  protected:
    /// The finite element space / basis associated with the rows
    std::shared_ptr<RowBasis> rowBasis_;

    /// The finite element space / basis associated with the columns
    std::shared_ptr<ColBasis> colBasis_;

    /// Data backend
    Backend backend_;

    /// Dense matrix to store coefficients during \ref assemble()
    ElementMatrix elementMatrix_;

    /// List of operators associated to row/col node
    MatrixOperators<RowBasis,ColBasis,ElementMatrix> operators_;
  };

} // end namespace AMDiS

#include "DOFMatrixBase.inc.hpp"
