#pragma once

#include <amdis/Assembler.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/typetree/Traversal.hpp>
#include <amdis/utility/AssembleOperators.hpp>

namespace AMDiS {

template <class RB, class CB, class B>
void DOFMatrixBase<RB,CB,B>::
init(bool asmMatrix)
{
  backend_.init(*rowBasis_, *colBasis_, asmMatrix);

  auto const rowSize = rowBasis_->localView().maxSize();
  auto const colSize = colBasis_->localView().maxSize();
  elementMatrix_.resize(rowSize, colSize);
}


template <class RB, class CB, class B>
void DOFMatrixBase<RB,CB,B>::
finish(bool /*asmMatrix*/)
{
  backend_.finish();
}


template <class RB, class CB, class B>
void DOFMatrixBase<RB,CB,B>::
insert(RowLocalView const& rowLocalView, ColLocalView const& colLocalView,
       ElementMatrix const& elementMatrix)
{
  using std::abs;
  for (size_type i = 0; i < rowLocalView.size(); ++i) {
    size_type const row = flatMultiIndex(rowLocalView.index(i));
    for (size_type j = 0; j < colLocalView.size(); ++j) {
      if (abs(elementMatrix[i][j]) > threshold<double>) {
        size_type const col = flatMultiIndex(colLocalView.index(j));
        backend_.insert(row, col, elementMatrix[i][j]);
      }
    }
  }
}


template <class RB, class CB, class B>
  template <class ContextTag, class Expr, class RowTreePath, class ColTreePath>
void DOFMatrixBase<RB,CB,B>::
addOperator(ContextTag contextTag, Expr const& expr,
            RowTreePath row, ColTreePath col)
{
  static_assert( Concepts::PreTreePath<RowTreePath>,
      "row must be a valid treepath, or an integer/index-constant");
  static_assert( Concepts::PreTreePath<ColTreePath>,
      "col must be a valid treepath, or an integer/index-constant");

  auto i = child(rowBasis_->localView().tree(), makeTreePath(row));
  auto j = child(colBasis_->localView().tree(), makeTreePath(col));

  using LocalContext = typename ContextTag::type;
  using Traits = DefaultAssemblerTraits<LocalContext, ElementMatrix>;
  auto op = makeLocalOperator<LocalContext>(expr, rowBasis_->gridView());
  auto localAssembler = makeUniquePtr(makeAssembler<Traits>(std::move(op), i, j));

  operators_[i][j].push(contextTag, std::move(localAssembler));
}


template <class RB, class CB, class B>
void DOFMatrixBase<RB,CB,B>::
assemble(RowLocalView const& rowLocalView, ColLocalView const& colLocalView)
{
  elementMatrix_ = 0;
  auto const& gv = rowBasis_->gridView();
  auto const& element = rowLocalView.element();
  auto geometry = element.geometry();

  for_each_node(rowLocalView.tree(), [&](auto const& rowNode, auto) {
    for_each_node(colLocalView.tree(), [&](auto const& colNode, auto) {
      auto& matOp = operators_[rowNode][colNode];
      if (matOp) {
        matOp.bind(element, geometry);
        assembleOperators(gv, element, matOp, makeMatrixAssembler(rowNode, colNode, elementMatrix_));
        matOp.unbind();
      }
    });
  });

  insert(rowLocalView, colLocalView, elementMatrix_);
}


template <class RB, class CB, class B>
void DOFMatrixBase<RB,CB,B>::
assemble()
{
  auto rowLocalView = rowBasis_->localView();
  auto colLocalView = colBasis_->localView();

  init(true);
  for (auto const& element : elements(rowBasis_->gridView())) {
    rowLocalView.bind(element);
    if (rowBasis_ == colBasis_)
      assemble(rowLocalView, rowLocalView);
    else {
      colLocalView.bind(element);
      assemble(rowLocalView, colLocalView);
      colLocalView.unbind();
    }
    rowLocalView.unbind(element);
  }
  finish(true);
}


} // end namespace AMDiS
